#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <assert.h>
#include "mkdat.h"

#define PROGRAM_NAME "mkdat"
#define PROGRAM_VERSION "1.0"

extern FILE *yyin;

const char *input_filename = NULL;
const char *output_filename = NULL;

char *data_written;
int data_alloc;
int data_size;

int debug_mode  = 0;
int swap_endian = 0;
int force_short = 0;
int align_short = 2;
int align_int   = 2;
int align_long  = 2;

void handle_args(int argc, char **argv);

void write_align(int alignment);
void write_numeric_data(char *data, size_t size);

void write_label(int identifier, unsigned char flags);
void write_reference(int identifier);
void write_shuffle_reference(int index);
int string_table_add(const char *str);
void write_string(int index);
void relocate(void);
void output(void);
void debug_output(void);

int get_identifier(const char *id);
void free_identifiers(void);
const char *identifier_string(int id);

void free_tables(void);

int get_constant(int identifier);
void define_constant(int identifier, int value);

void start_diversion(void);
void finish_diversion(void);

static void init_output(void);
void write_data(char *data, size_t size);

//{{{
int main(int argc, char **argv)
{
	init_output();
	
	atexit(free_identifiers);
	atexit(free_tables);
	
	handle_args(argc, argv);
	
	relocate();
	if(debug_mode)
		debug_output();
	else
		output();
	
	return 0;
} 
//}}}

//{{{
void handle_args(int argc, char **argv)
{
	int i;
	int num_inputs = 0;
	FILE *fin;
	const char *input_names[32];
	
	for(i=1; i<argc; i++)
	{
		if( !strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") ) {
			printf("Usage: %s [-a] [-d] [-e] [-h] [-s] [-v] [input] [-o output]\n"
			       "\n"
				   "  -a, --align     Align ints and longs on 4-byte boundaries\n"
				   "  -d, --debug     Print hex dump to stdout instead of binary to a file\n"
			       "  -e, --endian    Reverse byte order of output\n"
			       "  -h, --help      Show this page\n"
				   "  -s, --short     Ints are two bytes\n"
			       "  -v, --version   Show version information\n"
				   "\n",
			       argv[0]);
			exit(0);
		}
		else if( !strcmp(argv[i], "-v") || !strcmp(argv[i], "--version") )
		{
			printf(PROGRAM_NAME ", " PROGRAM_VERSION "\n"
			       "by Jim Babcock. Distributed under GNU Lesser General Public License (LGPL).\n"
			       "See the file COPYING for details.\n");
			exit(0);
		}
		else if( !strcmp(argv[i], "-e") || !strcmp(argv[i], "--endian") )
		{
			swap_endian = 1;
		}
		else if( !strcmp(argv[i], "-d") || !strcmp(argv[i], "--debug") )
		{
			debug_mode = 1;
		}
		else if( !strcmp(argv[i], "-a") || !strcmp(argv[i], "--align") )
		{
			align_int = 4;
			align_long = 4;
		}
		else if( !strcmp(argv[i], "-s") || !strcmp(argv[i], "--short") )
		{
			force_short = 1;
		}
		else if( !strcmp(argv[i], "-o") )
		{
			i++;
			if(i >= argc) {
				fprintf(stderr, "-o option requires an argument.\n");
				exit(ERROR_INVOCATION);
			}
			if(output_filename) {
				fprintf(stderr, "Error: Can only have one output location.\n");
				exit(ERROR_INVOCATION);
			}
			output_filename = argv[i];
		}
		else
		{
			// Input file
			num_inputs++;
			if(num_inputs>32) {
				fprintf(stderr, "Too many input files (limit 32)");
				exit(ERROR_INVOCATION);
			}
			input_names[num_inputs-1] = argv[i];
		}
	}
	
	if(num_inputs == 0)
	{
		fprintf(stderr, "Error: No inputs specified.\n");
		exit(ERROR_INVOCATION);
	}
	
	for(i=0; i<num_inputs; i++)
	{
		input_filename = input_names[i];
		fin = fopen(input_filename, "r");
		
		if(!fin) {
			fprintf(stderr, "%s: Error opening file\n", input_filename);
			exit(ERROR_FILES);
		}
		
		yyin = fin;
		yyparse();
		
		fclose(fin);
	}
	
	if(output_filename == NULL && debug_mode == 0)
	{
		fprintf(stderr, "Error: No output location specified.\n");
		exit(ERROR_INVOCATION);
	}
}
//}}}
//{{{
void output(void)
{
	FILE *fout;
	
	fout = fopen(output_filename, "wb");
	
	if(!fout)
	{
		fprintf(stderr, "Error opening output file %s.\n", output_filename);
		exit(ERROR_FILES);
	}
	
	fwrite(data_written, data_size, 1, fout);
	
	fclose(fout);
}
//}}}
//{{{
void put_hex(int n)
{
	if(n<10) putchar(n+'0');
	else putchar(n-10+'A');
}
//}}}
//{{{
void debug_output(void)
{
	int i;
	for(i=0; i<data_size; i++)
	{
		if(i%16 == 0 && i>0)
			putchar('\n');
		else if(i>0)
			putchar(' ');
		put_hex( (data_written[i]>>4)&0x0F );
		put_hex( (data_written[i]&0x0F) );
	}
	putchar('\n');
}
//}}}

//{{{
typedef struct filelink
{
	unsigned char flags;
	unsigned char file;
	unsigned short offset;
} filelink;
//}}}
//{{{
typedef struct label_info
{
	int identifier;
	filelink ptr;
} label_info;
//}}}
label_info *label_table = NULL;
int label_alloc = 0;
int label_num = 0;
//{{{
void write_label(int identifier, unsigned char flags)
{
	int i;
	label_info addition;
	
	// If this is the first time, allocate space for label info
	init_table((void**)&label_table, &label_alloc, 64, sizeof(label_info));
	
	// Check for duplicate labels
	for(i=0; i<label_num; i++)
	{
		if(label_table[i].identifier == identifier) {
			warn_dup_label(identifier_string(identifier));
			return;
		}
	}
	
	// Allocate more space if necessary
	expand_table((void**)&label_table, &label_alloc, label_num+1, sizeof(label_info));
	
	// Add label
	addition.identifier = identifier;
	addition.ptr.flags = flags;
	addition.ptr.file = 0;
	addition.ptr.offset = data_size;
	numeric_bytes((void*)&addition.ptr.offset, sizeof(addition.ptr.offset));
	label_table[label_num++] = addition;
}
//}}}

//{{{
typedef struct relocation_info
{
	int target;
	int location;
} relocation_info;
//}}}
relocation_info *relocation_table = NULL;
int relocation_alloc = 0;
int relocation_num = 0;
//{{{
void write_reference(int identifier)
{
	relocation_info addition;
	filelink placeholder;
	
	init_table((void**)&relocation_table, &relocation_alloc, 64, sizeof(relocation_info));
	expand_table((void**)&relocation_table, &relocation_alloc, relocation_num+1, sizeof(relocation_info));
	
	// Add relocation entry
	addition.target = identifier;
	addition.location = data_size;
	relocation_table[relocation_num++] = addition;
	
	placeholder.flags = 0;
	placeholder.file = 0;
	placeholder.offset = 0;
	write_data((void*)&placeholder, sizeof(placeholder));
}
//}}}
//{{{
void write_shuffle_reference(int index)
{
	filelink output;
	
	output.flags = 2;
	output.file = 0;
	output.offset = index;
	numeric_bytes((void*)&output.offset, sizeof(output.offset));
	write_data((void*)&output, sizeof(output));
}
//}}}

char **string_table = NULL;
int stringtab_alloc = 0;
int stringtab_num   = 0;
//{{{
// Input is a tokenized string, meaning it still has quote marks around it and
// escape sequences inside. Input length is guaranteed to be at least 2
// (two matching quotes)
int string_table_add(const char *str)
{
	int write_pos = 0;
	const char *read_pos;
	
	init_table((void**)&string_table, &stringtab_alloc, 64, sizeof(char*));
	expand_table((void**)&string_table, &stringtab_alloc, stringtab_num+1, sizeof(char*));
	
	string_table[stringtab_num] = malloc(strlen(str) - 1); // +1 for terminator, -2 for quote marks
	
	for(read_pos = str+1; *read_pos != '\0' && read_pos[1] != '\0'; read_pos++)
	{
		if(*read_pos == '\\') {
			read_pos++;
			string_table[stringtab_num][write_pos++] = parse_escape(&read_pos);
			read_pos--;
		} else {
			string_table[stringtab_num][write_pos++] = *read_pos;
		}
	}
	string_table[stringtab_num][write_pos] = '\0';
	
	stringtab_num++;
	
	
	return stringtab_num - 1;
}
//}}}
//{{{
void write_string(int index)
{
	int len;
	
	if(index > stringtab_num || index < 0)
	{
		fprintf(stderr, "Invalid index: %i\n", index);
		error_internal_param();
	}
	
	len = strlen(string_table[index]) + 1;
	write_data(string_table[index], len);
}
//}}}

//{{{
void relocate(void)
{
	int i, j;
	
	for(i=0; i<relocation_num; i++)
	{
		for(j=0; j<label_num; j++)
		{
			if(relocation_table[i].target == label_table[j].identifier)
			{
				memcpy(data_written + relocation_table[i].location, &label_table[j].ptr, sizeof(filelink));
				break;
			}
		}
		if(j==label_num)
			warn_undefined_reference(identifier_string(relocation_table[i].target));
	}
}
//}}}
//{{{
void write_val(int type, int val)
{
	unsigned long  val_ulong;
	  signed long  val_slong;
	unsigned int   val_uint;
	  signed int   val_sint;
	unsigned short val_ushort;
	  signed short val_sshort;
	unsigned char  val_uchar;
	  signed char  val_schar;
	
	if(type == type_uint && force_short) type = type_ushort;
	if(type == type_sint && force_short) type = type_sshort;
	
	switch(type)
	{
		case type_ulong:
			val_ulong = val;
			write_align(align_int);
			write_numeric_data((void*)&val_ulong, sizeof(val_ulong));
			break;
		case type_slong:
			val_slong = val;
			write_align(align_int);
			write_numeric_data((void*)&val_slong, sizeof(val_slong));
			break;
		case type_uint:
			val_uint = val;
			write_align(align_int);
			write_numeric_data((void*)&val_uint, sizeof(val_uint));
			break;
		case type_sint:
			val_sint = val;
			write_align(align_int);
			write_numeric_data((void*)&val_sint, sizeof(val_sint));
			break;
		case type_ushort:
			if(val > USHRT_MAX || val < 0)
				warn_range(val);
			
			val_ushort = val;
			write_align(align_short);
			write_numeric_data((void*)&val_ushort, sizeof(val_ushort));
			break;
		case type_sshort:
			if(val > SHRT_MAX || val < SHRT_MIN)
				warn_range(val);
			
			val_sshort = val;
			write_align(align_short);
			write_numeric_data((void*)&val_sshort, sizeof(val_sshort));
			break;
		case type_uchar:
			if(val > UCHAR_MAX || val < 0)
				warn_range(val);
			
			val_uchar = val;
			write_numeric_data((void*)&val_uchar, sizeof(val_uchar));
			break;
		case type_schar:
			if(val > SCHAR_MAX || val < SCHAR_MIN)
				warn_range(val);
			
			val_schar = val;
			write_numeric_data((void*)&val_schar, sizeof(val_schar));
			break;
	}
}
//}}}
//{{{
// Align the write cursor so that the current offset is a multiple of
// alignment. Zeroes are written as needed for padding.
void write_align(int alignment)
{
	char zero = 0;
	
	while(data_size % alignment)
	{
		write_data((void*)&zero, 1);
	}
}
//}}}
//{{{
void write_even(void)
{
	write_align(2);
}
//}}}
//{{{
// Write data with possibly reversed byte-order (if swap_endian is set)
void write_numeric_data(char *data, size_t size)
{
	numeric_bytes(data, size);
	
	write_data(data, size);
}
//}}}

//{{{
typedef struct identifier_info
{
	char *identifier;
	int constant_value;
	char is_constant;
} identifier_info;
//}}}
identifier_info *identifier_table = NULL;
int identifier_alloc = 0;
int identifier_num = 0;

//{{{
// Get the numeric value representing an identifier, or assign one if the
// identifier does not have a numeric value assigned. Values are handed out
// sequentially, and the same string will always get the same result.
int get_identifier(const char *id)
{
	int i;
	
	// If this is the first time, allocate space for the identifier table
	init_table((void**)&identifier_table, &identifier_alloc, 64, sizeof(identifier_info));
	
	// Check for previous use of this identifier
	for(i=0; i<identifier_num; i++)
	{
		if( strcmp(identifier_table[i].identifier, id) == 0 )
			return i;
	}
	
	// Enlarge the table if necessary
	expand_table((void**)&identifier_table, &identifier_alloc, identifier_num + 1, sizeof(identifier_info));
	
	// Add this identifier to the table
	identifier_table[identifier_num].identifier = strdup(id);
	identifier_table[identifier_num].constant_value = 0;
	identifier_table[identifier_num].is_constant = 0;
	
	return identifier_num++;
}
//}}}
//{{{
void free_identifiers(void)
{
	int i;
	
	if(!identifier_table)
		return;
	
	for(i=0; i<identifier_num; i++)
	{
		if(identifier_table[i].identifier)
		{
			free(identifier_table[i].identifier);
			identifier_table[i].identifier = NULL;
		}
	}
	
	identifier_num = 0;
}
//}}}
//{{{
const char *identifier_string(int id)
{
	return identifier_table[id].identifier;
}
//}}}

//{{{
void free_tables(void)
{
	if(data_written) {
		free(data_written);
		data_written = NULL;
	}
	if(label_table) {
		free(label_table);
		label_table = NULL;
	}
	if(relocation_table) {
		free(relocation_table);
		relocation_table = NULL;
	}
	if(identifier_table) {
		free(identifier_table);
		identifier_table = NULL;
	}
}
//}}}

//{{{
int get_constant(int identifier)
{
	if(!identifier_table[identifier].is_constant)
		error_undefined_identifier(identifier_string(identifier));
	
	return identifier_table[identifier].constant_value;
}
//}}}
//{{{
void define_constant(int identifier, int value)
{
	if(identifier_table[identifier].is_constant)
		warn_multiple_definition(identifier_string(identifier));
	
	identifier_table[identifier].is_constant = 1;
	identifier_table[identifier].constant_value = value;
}
//}}}

//{{{
void start_diversion(void)
{
}
//}}}
//{{{
void finish_diversion(void)
{
}
//}}}

//{{{
static void init_output(void)
{
	init_table((void**)&data_written, &data_alloc, 1024, 1);
}
//}}}
//{{{
void write_data(char *data, size_t size)
{
	expand_table((void**)&data_written, &data_alloc, data_size + size, 1);
	
	memcpy(data_written+data_size, data, size);
	data_size += size;
}
//}}}


